# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Define universal constants
"""
#: Gravitational constant :math:`G` in SI units (:math:`m^3 kg^{-1} s^{-2}`)
GRAVITATIONAL_CONST = 6.6743e-11

#: Vacuum magnetic permeability :math:`\mu_0` in SI units (:math:`NA^{-2}`)
VACUUM_MAGNETIC_PERMEABILITY = 1.25663706212e-06



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Magnetic forward modelling functions for rectangular prisms
"""
import numpy as np
from numba import jit

from ..constants import VACUUM_MAGNETIC_PERMEABILITY
from ._kernels import (
    kernel_ee,
    kernel_eee,
    kernel_een,
    kernel_eeu,
    kernel_en,
    kernel_enn,
    kernel_enu,
    kernel_eu,
    kernel_euu,
    kernel_nn,
    kernel_nnn,
    kernel_nnu,
    kernel_nu,
    kernel_nuu,
    kernel_uu,
    kernel_uuu,
)
from ._utils import (
    is_interior_point,
    is_point_on_east_face,
    is_point_on_edge,
    is_point_on_north_face,
    is_point_on_top_face,
)


@jit(nopython=True)
def magnetic_field(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Magnetic field due to a rectangular prism

    Returns the three components of the magnetic field due to a single
    rectangular prism on a single computation point.

    .. note::

        Use this function when all the three component of the magnetic fields
        are needed. Running this function is faster than computing each
        component separately. Use one of :func:`magnetic_e`,
        :func:`magnetic_n`, :func:`magnetic_u` if you need only one of them.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_e, b_n, b_u : float
        Easting, northing and upward component of the magnetic field generated
        by the prism on the observation point in :math:`\text{T}`.
        It will be ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices, prism edges or interior points.

    Notes
    -----
    Consider an observation point :math:`\mathbf{p}` and a prism :math:`R` with
    a magnetization vector :math:`\mathbf{M}`. The magnetic field
    :math:`\mathbf{B}` it generates on the observation point :math:`\mathbf{p}`
    is defined as:

    .. math::

        \mathbf{B}(\mathbf{p}) =
            - \frac{\mu_0}{4\pi} \nabla_\mathbf{p}
            \left[
            \int\limits_R \mathbf{M} \cdot \nabla_\mathbf{q}
            \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)
            dv
            \right]

    Since the magnetization vector is constant inside the boundaries of the
    prism, we can write the easting component of :math:`\mathbf{B}` as:

    .. math::

        B_x(\mathbf{p}) =
            - \frac{\mu_0}{4\pi}
            \left[
                M_x \int\limits_R
                \frac{\partial}{\partial x_p}
                \left[
                \frac{\partial}{\partial x_q}
                \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)
                \right]
                dv
                +
                M_y \int\limits_R
                \frac{\partial}{\partial x_p}
                \left[
                \frac{\partial}{\partial y_q}
                \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)
                \right]
                dv
                +
                M_z \int\limits_R
                \frac{\partial}{\partial x_p}
                \left[
                \frac{\partial}{\partial z_q}
                \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)
                \right]
                dv
            \right]

    where :math:`M_x`, :math:`M_y` and :math:`M_z` are the components of the
    magnetization vector. The other components can be expressed in an analogous
    way.

    It can be proved that

    .. math::

        \frac{\partial}{\partial x_q}
        \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)
        =
        - \frac{\partial}{\partial x_p}
        \left( \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert} \right)

    and that it also holds for the two other directions. Therefore, we can
    rewrite :math:`B_x` as:

    .. math::

        B_x(\mathbf{p}) =
            + \frac{\mu_0}{4\pi}
            \left[
                M_x
                \frac{\partial^2}{\partial x_p^2}
                \int\limits_R
                \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
                dv
                +
                M_y
                \frac{\partial^2}{\partial x_p \partial y_p}
                \int\limits_R
                \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
                dv
                +
                M_z
                \frac{\partial^2}{\partial x_p \partial z_p}
                \int\limits_R
                \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
                dv
            \right]

    Solutions to each one of the integrals in the previous equation and their
    second derivatives are given by [Nagy2000]_.

    Following [Oliveira2015]_ we can define a symmetrical 3x3 matrix
    :math:`\mathbf{U}` whose elements are the second derivatives of the
    previous integrals, such as:

    .. math::

        u_{ij} =
            \frac{\partial^2}{\partial i \partial j}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i, j \in \{x, y, z\}`.

    We can then express the magnetic field :math:`\mathbf{B}(\mathbf{p})`
    generated by the prism in a compact form:

    .. math::

        \mathbf{B}(\mathbf{p}) = \frac{\mu_0}{4\pi} \mathbf{U} \cdot \mathbf{M}

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return (np.nan, np.nan, np.nan)
    # Initialize magnetic field vector components
    b_e, b_n, b_u = 0.0, 0.0, 0.0
    # Iterate over the vertices of the prism
    for i in range(2):
        # Compute shifted easting coordinate
        if i == 0:
            shift_east = prism_east - easting
        else:
            shift_east = prism_west - easting
        shift_east_sq = shift_east**2
        for j in range(2):
            # Compute shifted northing coordinate
            if j == 0:
                shift_north = prism_north - northing
            else:
                shift_north = prism_south - northing
            shift_north_sq = shift_north**2
            for k in range(2):
                # Compute shifted upward coordinate
                if k == 0:
                    shift_upward = prism_top - upward
                else:
                    shift_upward = prism_bottom - upward
                shift_upward_sq = shift_upward**2
                # Compute the radius
                radius = np.sqrt(shift_east_sq + shift_north_sq + shift_upward_sq)
                # Compute all kernel tensor components for the current vertex
                k_ee = kernel_ee(shift_east, shift_north, shift_upward, radius)
                k_nn = kernel_nn(shift_east, shift_north, shift_upward, radius)
                k_uu = kernel_uu(shift_east, shift_north, shift_upward, radius)
                k_en = kernel_en(shift_east, shift_north, shift_upward, radius)
                k_eu = kernel_eu(shift_east, shift_north, shift_upward, radius)
                k_nu = kernel_nu(shift_east, shift_north, shift_upward, radius)
                # Get the sign of this terms based on the current vertex
                sign = (-1) ** (i + j + k)
                # Compute the dot product between the kernel tensor and the
                # magnetization vector of the prism
                b_e += sign * (
                    magnetization_east * k_ee
                    + magnetization_north * k_en
                    + magnetization_up * k_eu
                )
                b_n += sign * (
                    magnetization_east * k_en
                    + magnetization_north * k_nn
                    + magnetization_up * k_nu
                )
                b_u += sign * (
                    magnetization_east * k_eu
                    + magnetization_north * k_nu
                    + magnetization_up * k_uu
                )
    # Add 4 pi to Be if computing on the eastmost face, to correctly evaluate
    # the limit approaching from outside (approaching from the east)
    if is_point_on_east_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_e += 4 * np.pi * magnetization_east
    # Add 4 pi to Bn if computing on the northmost face, to correctly evaluate
    # the limit approaching from outside (approaching from the north)
    if is_point_on_north_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_n += 4 * np.pi * magnetization_north
    # Add 4 pi to Bu if computing on the north face, to correctly evaluate the
    # limit approaching from outside (approaching from the top)
    if is_point_on_top_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_u += 4 * np.pi * magnetization_up
    c_m = VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi
    b_e *= c_m
    b_n *= c_m
    b_u *= c_m
    return b_e, b_n, b_u


@jit(nopython=True)
def magnetic_e(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Easting component of the magnetic field due to a prism

    Returns the easting component of the magnetic field due to a single
    rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_e : float
        Easting component of the magnetic field generated by the prism
        on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the easting component of the magnetic field
    :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism :math:`R`
    with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_x(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xx} + M_y u_{xy} + M_z u_{xz} \right)

    Where :math:`u_{ij}` are:

    .. math::

        u_{ij} =
            \frac{\partial^2}{\partial i \partial j}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j \in \{x, y, z\}`.
    Solutions of the second derivatives of these integrals are given by
    [Nagy2000]_:

    .. math::

        u_{xx} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            - \arctan \left( \frac{yz}{xr} \right)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{xy} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (z + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{xz} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (y + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic field vector component
    b_e = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_ee,
        kernel_en,
        kernel_eu,
    )
    # Add 4 pi to Be if computing on the eastmost face, to correctly evaluate
    # the limit approaching from outside (approaching from the east)
    if is_point_on_east_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_e += 4 * np.pi * magnetization_east
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_e


@jit(nopython=True)
def magnetic_n(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Northing component of the magnetic field due to a prism

    Returns the northing component of the magnetic field due to a single
    rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_n : float
        Northing component of the magnetic field generated by the prism
        on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the northing component of the magnetic field
    :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism :math:`R`
    with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_y(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xy} + M_y u_{yy} + M_z u_{yz} \right)

    Where :math:`u_{ij}` are:

    .. math::

        u_{ij} =
            \frac{\partial^2}{\partial i \partial j}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j \in \{x, y, z\}`.
    Solutions of the second derivatives of these integrals are given by
    [Nagy2000]_:

    .. math::

        u_{xy} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (z + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{yy} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            - \arctan \left( \frac{xz}{yr} \right)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{yz} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (x + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic field vector component
    b_n = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_en,
        kernel_nn,
        kernel_nu,
    )
    # Add 4 pi to Bn if computing on the northmost face, to correctly evaluate
    # the limit approaching from outside (approaching from the north)
    if is_point_on_north_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_n += 4 * np.pi * magnetization_north
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_n


@jit(nopython=True)
def magnetic_u(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Upward component of the magnetic field due to a prism

    Returns the upward component of the magnetic field due to a single
    rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_u : float
        Upward component of the magnetic field generated by the prism
        on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the upward component of the magnetic field
    :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism :math:`R`
    with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_z(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xz} + M_y u_{yz} + M_z u_{zz} \right)

    Where :math:`u_{ij}` are:

    .. math::

        u_{ij} =
            \frac{\partial^2}{\partial i \partial j}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j \in \{x, y, z\}`.
    Solutions of the second derivatives of these integrals are given by
    [Nagy2000]_:

    .. math::

        u_{xz} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (y + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{yz} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            \ln (x + r)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}
        \\
        u_{zz} &=
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
            - \arctan \left( \frac{xy}{zr} \right)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic field vector component
    b_u = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_eu,
        kernel_nu,
        kernel_uu,
    )
    # Add 4 pi to Bu if computing on the north face, to correctly evaluate the
    # limit approaching from outside (approaching from the top)
    if is_point_on_top_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        b_u += 4 * np.pi * magnetization_up
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_u


@jit(nopython=True)
def magnetic_ee(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Easting derivative of the easting component of the magnetic field.

    Returns the easting derivative of the easting component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_ee : float
        Easting derivative of the easting component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the easting derivative of the easting component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{xx}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xxx} + M_y u_{xxy} + M_z u_{xxz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_ee = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_eee,
        kernel_een,
        kernel_eeu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_ee


@jit(nopython=True)
def magnetic_nn(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Northing derivative of the northing component of the magnetic field.

    Returns the northing derivative of the northing component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_nn : float
        Northing derivative of the northing component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the northing derivative of the northing component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{yy}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xyy} + M_y u_{yyy} + M_z u_{yyz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_nn = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_enn,
        kernel_nnn,
        kernel_nnu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_nn


@jit(nopython=True)
def magnetic_uu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Upward derivative of the upward component of the magnetic field.

    Returns the upward derivative of the upward component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_uu : float
        Upward derivative of the upward component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in
        a singular point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the upward derivative of the upward component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{zz}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xzz} + M_y u_{yzz} + M_z u_{zzz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_uu = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_euu,
        kernel_nuu,
        kernel_uuu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_uu


@jit(nopython=True)
def magnetic_en(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Northing derivative of the easting component of the magnetic field.

    Returns the northing derivative of the easting component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_en : float
        Northing derivative of the easting component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the northing derivative of the easting component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{xy}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xxy} + M_y u_{xyy} + M_z u_{xyz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_eu`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_en = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_een,
        kernel_enn,
        kernel_enu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_en


@jit(nopython=True)
def magnetic_eu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Upward derivative of the easting component of the magnetic field.

    Returns the upward derivative of the easting component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_eu : float
        Upward derivative of the easting component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the northing derivative of the easting component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{xz}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xxz} + M_y u_{xyz} + M_z u_{xzz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_nu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_eu = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_eeu,
        kernel_enu,
        kernel_euu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_eu


@jit(nopython=True)
def magnetic_nu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
):
    r"""
    Upward derivative of the northing component of the magnetic field.

    Returns the upward derivative of the northing component of the magnetic
    field due to a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east : float
        The East component of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    magnetization_north : float
        The North component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.
    magnetization_up : float
        The upward component of the magnetization vector of the prism. Must be
        in :math:`A m^{-1}`.

    Returns
    -------
    b_nu : float
        Upward derivative of the northing component of the magnetic field
        generated by the prism on the observation point in :math:`\text{T}`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices, prism edges or interior points.

    Notes
    -----
    Computes the northing derivative of the easting component of the magnetic
    field :math:`\mathbf{B}(\mathbf{p})` generated by a rectangular prism
    :math:`R` with a magnetization vector :math:`M` on the observation point
    :math:`\mathbf{p}` as follows:

    .. math::

        B_{yz}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left( M_x u_{xyz} + M_y u_{yyz} + M_z u_{yzz} \right)

    Where :math:`u_{ijk}` are:

    .. math::

        u_{ijk} =
            \frac{\partial^3}{\partial i \partial j \partial k}
            \int\limits_R
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert}
            dv

    with :math:`i,j,k \in \{x, y, z\}`.

    References
    ----------
    - [Blakely1995]_
    - [Oliveira2015]_
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_

    See Also
    --------
    :func:`choclo.prism.magnetic_field`
    :func:`choclo.prism.magnetic_e`
    :func:`choclo.prism.magnetic_n`
    :func:`choclo.prism.magnetic_u`
    :func:`choclo.prism.magnetic_ee`
    :func:`choclo.prism.magnetic_nn`
    :func:`choclo.prism.magnetic_uu`
    :func:`choclo.prism.magnetic_en`
    :func:`choclo.prism.magnetic_eu`
    """
    # Check if observation point falls in a singular point
    if is_point_on_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_interior_point(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Compute magnetic gradiometry component
    b_nu = _calculate_component(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        magnetization_east,
        magnetization_north,
        magnetization_up,
        kernel_enu,
        kernel_nnu,
        kernel_nuu,
    )
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * b_nu


@jit(nopython=True)
def _calculate_component(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    magnetization_east,
    magnetization_north,
    magnetization_up,
    kernel_i,
    kernel_j,
    kernel_k,
):
    r"""
    Calculate field component for a single prism and observation point.

    Evaluate the provided kernels on the shifted coordinates of prism vertices
    to compute the magnetic field component given a magnetization vector of the
    prism.

    Parameters
    ----------
    easting, northing, upward : float
        Coordinates of the observation point. Must be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    magnetization_east, magnetization_north, magnetization_up : float
        The components of the magnetization vector of the prism. Must be in
        :math:`A m^{-1}`.
    kernel_i, kernel_j, kernel_k : callables
        Kernel functions to be evaluated on each vertex of the prism.

    Returns
    -------
    float

    Notes
    -----
    Given the kernels :math:`k_i(\hat{x}, \hat{y}, \hat{z})`,
    :math:`k_j(\hat{x}, \hat{y}, \hat{z})`, and :math:`k_k(\hat{x}, \hat{y},
    \hat{z})`; a prism provided by its boundaries :math:`x_1`, :math:`x_2`,
    :math:`y_1`, :math:`y_2`, :math:`z_1`, and :math:`z_2`; a magnetization
    vector :math:`\mathbf{M}=(M_x, M_y, M_z)`; and an observation point
    :math:`\mathbf{p}=(x, y, z)`, this function returns:

    .. math::

        M_x u_x(x, y, z) + M_y u_y(x, y, z) + M_z u_z(x, y, z),

    where

    .. math::

        u_x(x, y, z) =
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
              k_i(\hat{x}, \hat{y}, \hat{z})
            \Bigg\rvert_{x_1 - x}^{x_2 - x}
            \Bigg\rvert_{y_1 - y}^{y_2 - y}
            \Bigg\rvert_{z_1 - z}^{z_2 - z}

    .. math::

        u_y(x, y, z) =
            \Bigg\lvert\Bigg\lvert\Bigg\lvert
              k_j(\hat{x}, \hat{y}, \hat{z})
            \Bigg\rvert_{x_1 - x}^{x_2 - x}
            \Bigg\rvert_{y_1 - y}^{y_2 - y}
            \Bigg\rvert_{z_1 - z}^{z_2 - z}

    .. math::

        u_z(x, y, z) =
          \Bigg\lvert\Bigg\lvert\Bigg\lvert
            k_k(\hat{x}, \hat{y}, \hat{z})
          \Bigg\rvert_{x_1 - x}^{x_2 - x}
          \Bigg\rvert_{y_1 - y}^{y_2 - y}
          \Bigg\rvert_{z_1 - z}^{z_2 - z}
    """
    result = 0.0
    # Iterate over the vertices of the prism
    for i in range(2):
        # Compute shifted easting coordinate
        if i == 0:
            shift_east = prism_east - easting
        else:
            shift_east = prism_west - easting
        shift_east_sq = shift_east**2
        for j in range(2):
            # Compute shifted northing coordinate
            if j == 0:
                shift_north = prism_north - northing
            else:
                shift_north = prism_south - northing
            shift_north_sq = shift_north**2
            for k in range(2):
                # Compute shifted upward coordinate
                if k == 0:
                    shift_upward = prism_top - upward
                else:
                    shift_upward = prism_bottom - upward
                shift_upward_sq = shift_upward**2
                # Compute the radius
                radius = np.sqrt(shift_east_sq + shift_north_sq + shift_upward_sq)
                # Compute kernel tensor components for the current vertex
                k_e = kernel_i(shift_east, shift_north, shift_upward, radius)
                k_n = kernel_j(shift_east, shift_north, shift_upward, radius)
                k_u = kernel_k(shift_east, shift_north, shift_upward, radius)
                # Compute b_en using the dot product between the kernel tensor
                # and the magnetization vector of the prism
                result += (-1) ** (i + j + k) * (
                    magnetization_east * k_e
                    + magnetization_north * k_n
                    + magnetization_up * k_u
                )
    return result



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Test gravity forward modelling functions for point sources
"""
import numpy as np
import numpy.testing as npt
import pytest

from ..point import (
    gravity_e,
    gravity_ee,
    gravity_en,
    gravity_eu,
    gravity_n,
    gravity_nn,
    gravity_nu,
    gravity_pot,
    gravity_u,
    gravity_uu,
)
from .utils import NUMBA_IS_DISABLED


@pytest.fixture(name="sample_point_source")
def fixture_sample_point_source():
    """
    Return a sample point source
    """
    return (3.7, 4.3, -5.8)


@pytest.fixture(name="sample_mass")
def fixture_sample_mass():
    """
    Return the mass for the sample point source
    """
    return 9e4


@pytest.fixture(name="sample_coordinate")
def fixture_sample_coordinate():
    """
    Define a sample observation point
    """
    return (16.7, 13.2, 7.8)


class TestSymmetryPotential:
    """
    Test the symmetry of gravity potential due to a point source
    """

    def build_points_in_sphere(self, radius, point):
        """
        Generate a set of observation points located in a sphere of radius
        ``radius`` and centered around the ``point``.
        """
        longitude = np.arange(0, 2 * np.pi, np.pi / 8)
        latitude = np.arange(-np.pi / 2, np.pi / 2, np.pi / 8)
        longitude, latitude = np.meshgrid(longitude, latitude)
        x = radius * np.cos(longitude) * np.cos(latitude)
        y = radius * np.sin(longitude) * np.cos(latitude)
        z = radius * np.sin(latitude)
        x += point[0]
        y += point[1]
        z += point[2]
        return tuple(coord.ravel() for coord in (x, y, z))

    def test_symmetry_on_sphere(self, sample_point_source, sample_mass):
        """
        Test the symmetry of the gravity potential in points of a sphere
        """
        radius = 3.5
        observation_points = self.build_points_in_sphere(radius, sample_point_source)
        # Evaluate gravity_pot on every observation point
        potential = tuple(
            gravity_pot(*coords, *sample_point_source, sample_mass)
            for coords in zip(*observation_points)
        )
        # Check if all values are equal
        npt.assert_allclose(potential[0], potential)

    def test_potential_between_two_spheres(self, sample_point_source, sample_mass):
        """
        Test the gravity potential between observation points in two spheres
        """
        radius_1, radius_2 = 3.5, 8.7
        sphere_1 = self.build_points_in_sphere(radius_1, sample_point_source)
        sphere_2 = self.build_points_in_sphere(radius_2, sample_point_source)
        # Evaluate kernel_pot on every observation point
        potential_1 = gravity_pot(*sphere_1, *sample_point_source, sample_mass)
        potential_2 = gravity_pot(*sphere_2, *sample_point_source, sample_mass)
        # Check if all values are equal
        npt.assert_allclose(potential_1, radius_2 / radius_1 * potential_2)

    @pytest.mark.skipif(not NUMBA_IS_DISABLED, reason="Numba is not disabled")
    def test_infinite_potential(self, sample_mass):
        """
        Test if we get an infinite kernel if the computation point is in the
        same location as the observation point.

        This test should be run only with Numba disabled.
        """
        point = (4.6, -8.9, -50.3)
        assert np.isinf(gravity_pot(*point, *point, sample_mass))

    @pytest.mark.skipif(NUMBA_IS_DISABLED, reason="Numba is disabled")
    def test_division_by_zero(self, sample_mass):
        """
        Test if we get a division by zero if the computation point is in the
        same location as the observation point.

        This test should be run only with Numba enabled.
        """
        point = (4.6, -8.9, -50.3)
        with pytest.raises(ZeroDivisionError):
            gravity_pot(*point, *point, sample_mass)


class TestSymmetryGravityE:
    """
    Test symmetry of the gravity_e
    """

    @pytest.fixture
    def coords_northing_upward_plane(self, sample_point_source):
        """
        Define set of observation points in the northing-upward plane
        """
        # Define northing and upward coordinates (avoid the zero, that's where
        # the sample_point_source is located)
        northing = np.linspace(-11, 11, 12)
        upward = np.linspace(-7, 7, 8)
        # Generate meshgrid
        northing, upward = np.meshgrid(northing, upward)
        # Compute easting
        easting = np.zeros_like(northing)
        # Add the coordinates of the sample point source
        easting += sample_point_source[0]
        northing += sample_point_source[1]
        upward += sample_point_source[2]
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_point_source):
        """
        Define two set of mirrored points across the northing-upward plane
        """
        # Define the northing and upward coordinates of the points
        northing = np.linspace(-11, 11, 27)
        upward = np.linspace(-38, 38, 23)
        # Define the easting coordinates for the first set
        easting_1 = np.linspace(1, 15, 15)
        # And mirror it for the second set
        easting_2 = -easting_1
        # Shift the coordinates to the sample_point_source
        easting_1 += sample_point_source[0]
        easting_2 += sample_point_source[0]
        northing += sample_point_source[1]
        upward += sample_point_source[2]
        return (easting_1, northing, upward), (easting_2, northing, upward)

    def test_zero_in_northing_upward(
        self, sample_point_source, coords_northing_upward_plane, sample_mass
    ):
        """
        Test if kernel_e is zero in the northing-upward plane
        """
        g_e = np.array(
            list(
                gravity_e(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_northing_upward_plane)
            )
        )
        assert (g_e <= 1e-30).all()

    def test_mirror_symmetry(self, sample_point_source, mirrored_points, sample_mass):
        """
        Test points with opposite easting coordinate
        """
        coords_1, coords_2 = mirrored_points
        g_e_1 = np.array(
            list(
                gravity_e(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_1)
            )
        )
        g_e_2 = np.array(
            list(
                gravity_e(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_2)
            )
        )
        npt.assert_allclose(g_e_1, -g_e_2)


class TestSymmetryGravityN:
    """
    Test symmetry of the gravity_n
    """

    @pytest.fixture
    def coords_easting_upward_plane(self, sample_point_source):
        """
        Define set of observation points in the easting-upward plane
        """
        # Define easting and upward coordinates (avoid the zero, that's where
        # the sample_point_source is located)
        easting = np.linspace(-11, 11, 12)
        upward = np.linspace(-7, 7, 8)
        # Generate meshgrid
        easting, upward = np.meshgrid(easting, upward)
        # Compute northing
        northing = np.zeros_like(easting)
        # Add the coordinates of the sample point source
        easting += sample_point_source[0]
        northing += sample_point_source[1]
        upward += sample_point_source[2]
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_point_source):
        """
        Define two set of mirrored points across the easting-upward plane
        """
        # Define the easting and upward coordinates of the points
        easting = np.linspace(-11, 11, 27)
        upward = np.linspace(-38, 38, 23)
        # Define the northing coordinates for the first set
        northing_1 = np.linspace(1, 15, 15)
        # And mirror it for the second set
        northing_2 = -northing_1
        # Shift the coordinates to the sample_point_source
        easting += sample_point_source[0]
        northing_1 += sample_point_source[1]
        northing_2 += sample_point_source[1]
        upward += sample_point_source[2]
        return (easting, northing_1, upward), (easting, northing_2, upward)

    def test_zero_in_easting_upward(
        self, sample_point_source, coords_easting_upward_plane, sample_mass
    ):
        """
        Test if kernel_e is zero in the northing-upward plane
        """
        g_n = np.array(
            list(
                gravity_n(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_easting_upward_plane)
            )
        )
        assert (g_n <= 1e-30).all()

    def test_mirror_symmetry(self, sample_point_source, mirrored_points, sample_mass):
        """
        Test points with opposite easting coordinate
        """
        coords_1, coords_2 = mirrored_points
        g_n_1 = np.array(
            list(
                gravity_n(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_1)
            )
        )
        g_n_2 = np.array(
            list(
                gravity_n(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_2)
            )
        )
        npt.assert_allclose(g_n_1, -g_n_2)


class TestSymmetryGravityU:
    """
    Test symmetry of the gravity_u
    """

    @pytest.fixture
    def coords_easting_northing_plane(self, sample_point_source):
        """
        Define set of observation points in the easting-northing plane
        """
        # Define easting and northing coordinates (avoid the zero, that's where
        # the sample_point_source is located)
        easting = np.linspace(-11, 11, 12)
        northing = np.linspace(-7, 7, 8)
        # Generate meshgrid
        easting, northing = np.meshgrid(easting, northing)
        # Compute upward
        upward = np.zeros_like(easting)
        # Add the coordinates of the sample point source
        easting += sample_point_source[0]
        northing += sample_point_source[1]
        upward += sample_point_source[2]
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_point_source):
        """
        Define two set of mirrored points across the easting-northing plane
        """
        # Define the easting and upward coordinates of the points
        easting = np.linspace(-11, 11, 27)
        northing = np.linspace(-38, 38, 23)
        # Define the northing coordinates for the first set
        upward_1 = np.linspace(1, 15, 15)
        # And mirror it for the second set
        upward_2 = -upward_1
        # Shift the coordinates to the sample_point_source
        easting += sample_point_source[0]
        northing += sample_point_source[1]
        upward_1 += sample_point_source[2]
        upward_2 += sample_point_source[2]
        return (easting, northing, upward_1), (easting, northing, upward_2)

    def test_zero_in_easting_northing(
        self, sample_point_source, coords_easting_northing_plane, sample_mass
    ):
        """
        Test if gravity_u is zero in the easting-upward plane
        """
        g_u = np.array(
            list(
                gravity_u(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_easting_northing_plane)
            )
        )
        assert (g_u <= 1e-30).all()

    def test_mirror_symmetry(self, sample_point_source, mirrored_points, sample_mass):
        """
        Test points with opposite upward coordinate
        """
        coords_1, coords_2 = mirrored_points
        g_u_1 = np.array(
            list(
                gravity_u(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_1)
            )
        )
        g_u_2 = np.array(
            list(
                gravity_u(*coords, *sample_point_source, sample_mass)
                for coords in zip(*coords_2)
            )
        )
        npt.assert_allclose(g_u_1, -g_u_2)


class TestAccelerationFiniteDifferences:
    """
    Test acceleration components against finite-differences approximations of
    the gravity potential
    """

    # Define percentage for the finite difference displacement
    delta_percentage = 1e-8

    # Define expected relative error tolerance in the comparisons
    rtol = 1e-5

    @pytest.fixture
    def finite_diff_gravity_e(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_e through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        easting_q, _, _ = sample_point_source
        # Compute a small increment in the easting coordinate
        d_easting = self.delta_percentage * (easting_p - easting_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p + d_easting, northing_p, upward_p)
        # Calculate g_e through finite differences
        g_e = (
            gravity_pot(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_pot(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_easting
        return g_e

    @pytest.fixture
    def finite_diff_gravity_n(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_n through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, northing_q, _ = sample_point_source
        # Compute a small increment in the easting coordinate
        d_northing = self.delta_percentage * (northing_p - northing_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + d_northing, upward_p)
        # Calculate g_n through finite differences
        g_n = (
            gravity_pot(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_pot(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_northing
        return g_n

    @pytest.fixture
    def finite_diff_gravity_u(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_u through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, _, upward_q = sample_point_source
        # Compute a small increment in the easting coordinate
        d_upward = self.delta_percentage * (upward_p - upward_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + d_upward)
        # Calculate g_u through finite differences
        g_u = (
            gravity_pot(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_pot(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_upward
        return g_u

    def test_gravity_e(
        self, sample_coordinate, sample_point_source, finite_diff_gravity_e, sample_mass
    ):
        """
        Test gravity_e against finite differences of the gravity_pot
        """
        npt.assert_allclose(
            finite_diff_gravity_e,
            gravity_e(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_n(
        self, sample_coordinate, sample_point_source, finite_diff_gravity_n, sample_mass
    ):
        """
        Test gravity_n against finite differences of the gravity_pot
        """
        npt.assert_allclose(
            finite_diff_gravity_n,
            gravity_n(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_u(
        self, sample_coordinate, sample_point_source, finite_diff_gravity_u, sample_mass
    ):
        """
        Test gravity_u against finite differences of the gravity_pot
        """
        npt.assert_allclose(
            finite_diff_gravity_u,
            gravity_u(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )


class TestTensorFiniteDifferences:
    """
    Test tensor gravity functions against finite-differences approximations of
    the gradient components
    """

    # Define percentage for the finite difference displacement
    delta_percentage = 1e-8

    # Define expected relative error tolerance in the comparisons
    rtol = 1e-5

    @pytest.fixture
    def finite_diff_gravity_ee(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_ee through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        easting_q, _, _ = sample_point_source
        # Compute a small increment in the easting coordinate
        d_easting = self.delta_percentage * (easting_p - easting_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p + d_easting, northing_p, upward_p)
        # Calculate g_ee through finite differences
        g_ee = (
            gravity_e(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_e(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_easting
        return g_ee

    @pytest.fixture
    def finite_diff_gravity_nn(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_nn through finite differences of the gravity_n
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, northing_q, _ = sample_point_source
        # Compute a small increment in the easting coordinate
        d_northing = self.delta_percentage * (northing_p - northing_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + d_northing, upward_p)
        # Calculate g_nn through finite differences
        g_nn = (
            gravity_n(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_n(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_northing
        return g_nn

    @pytest.fixture
    def finite_diff_gravity_uu(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_uu through finite differences of the gravity_u
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, _, upward_q = sample_point_source
        # Compute a small increment in the easting coordinate
        d_upward = self.delta_percentage * (upward_p - upward_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + d_upward)
        # Calculate g_uu through finite differences
        g_uu = (
            gravity_u(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_u(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_upward
        return g_uu

    @pytest.fixture
    def finite_diff_gravity_en(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_en through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, northing_q, _ = sample_point_source
        # Compute a small increment in the easting coordinate
        d_northing = self.delta_percentage * (northing_p - northing_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + d_northing, upward_p)
        # Calculate g_en through finite differences
        g_en = (
            gravity_e(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_e(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_northing
        return g_en

    @pytest.fixture
    def finite_diff_gravity_eu(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Compute gravity_eu through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, _, upward_q = sample_point_source
        # Compute a small increment in the easting coordinate
        d_upward = self.delta_percentage * (upward_p - upward_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + d_upward)
        # Calculate g_eu through finite differences
        g_eu = (
            gravity_e(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_e(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_upward
        return g_eu

    @pytest.fixture
    def finite_diff_gravity_nu(
        self, sample_coordinate, sample_point_source, sample_mass
    ):
        """
        Test gravity_nu against finite differences of the gravity_n
        """
        easting_p, northing_p, upward_p = sample_coordinate
        _, _, upward_q = sample_point_source
        # Compute a small increment in the easting coordinate
        d_upward = self.delta_percentage * (upward_p - upward_q)
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + d_upward)
        # Calculate g_nu through finite differences
        g_nu = (
            gravity_n(*shifted_coordinate, *sample_point_source, sample_mass)
            - gravity_n(*sample_coordinate, *sample_point_source, sample_mass)
        ) / d_upward
        return g_nu

    def test_gravity_ee(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_ee,
        sample_mass,
    ):
        """
        Test gravity_ee against finite differences of the gravity_e
        """
        npt.assert_allclose(
            finite_diff_gravity_ee,
            gravity_ee(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_nn(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_nn,
        sample_mass,
    ):
        """
        Test gravity_nn against finite differences of the gravity_n
        """
        npt.assert_allclose(
            finite_diff_gravity_nn,
            gravity_nn(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_uu(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_uu,
        sample_mass,
    ):
        """
        Test gravity_uu against finite differences of the gravity_u
        """
        npt.assert_allclose(
            finite_diff_gravity_uu,
            gravity_uu(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_en(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_en,
        sample_mass,
    ):
        """
        Test gravity_en against finite differences of the gravity_e
        """
        npt.assert_allclose(
            finite_diff_gravity_en,
            gravity_en(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_eu(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_eu,
        sample_mass,
    ):
        """
        Test gravity_eu against finite differences of the gravity_e
        """
        npt.assert_allclose(
            finite_diff_gravity_eu,
            gravity_eu(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )

    def test_gravity_nu(
        self,
        sample_coordinate,
        sample_point_source,
        finite_diff_gravity_nu,
        sample_mass,
    ):
        """
        Test gravity_nu against finite differences of the gravity_n
        """
        npt.assert_allclose(
            finite_diff_gravity_nu,
            gravity_nu(*sample_coordinate, *sample_point_source, sample_mass),
            rtol=self.rtol,
        )


class TestLaplacian:
    @pytest.fixture
    def sample_observation_points(self, sample_point_source):
        """
        Define a 3D grid of observation points around the sample point source.
        The grid doesn't contain an observation point located in the same
        location as the point source.
        """
        # Build the observation points
        easting = np.linspace(-10, 10, 21)
        northing = np.linspace(-10, 10, 21)
        upward = np.linspace(-10, 10, 21)
        easting, northing, upward = tuple(
            a.ravel() for a in np.meshgrid(easting, northing, upward)
        )
        # Remove the location of the sample point source
        point_source_location = (easting == 0) & (northing == 0) & (upward == 0)
        easting, northing, upward = tuple(
            a[np.logical_not(point_source_location)]
            for a in (easting, northing, upward)
        )
        # Shift the coordinates
        easting += sample_point_source[0]
        northing += sample_point_source[1]
        upward += sample_point_source[2]
        return (easting, northing, upward)

    @pytest.mark.parametrize("first_component", ("g_ee", "g_nn", "g_uu"))
    def test_laplacian(
        self,
        sample_observation_points,
        sample_point_source,
        sample_mass,
        first_component,
    ):
        """
        Test if diagonal tensor functions satisfy Laplace's equation
        """
        g_ee = np.array(
            [
                gravity_ee(e, n, u, *sample_point_source, sample_mass)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        g_nn = np.array(
            [
                gravity_nn(e, n, u, *sample_point_source, sample_mass)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        g_uu = np.array(
            [
                gravity_uu(e, n, u, *sample_point_source, sample_mass)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        # Set an atol to avoid getting failures when comparing values close to
        # zero
        atol = 1e-21
        rtol = 1e-7
        if first_component == "g_ee":
            npt.assert_allclose(-g_ee, g_nn + g_uu, rtol=rtol, atol=atol)
        if first_component == "g_nn":
            npt.assert_allclose(-g_nn, g_ee + g_uu, rtol=rtol, atol=atol)
        if first_component == "g_uu":
            npt.assert_allclose(-g_uu, g_ee + g_nn, rtol=rtol, atol=atol)



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Kernel functions for point sources
"""
from numba import jit


@jit(nopython=True)
def kernel_pot(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    The inverse of the distance between the two points

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_V(\mathbf{p}, \mathbf{q}) =
        \frac{
            1
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 1 / distance


@jit(nopython=True)
def kernel_e(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Easting component of the gradient of the inverse of the distance

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_x(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial x}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        - \frac{
            x_p - x_q
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return -(easting_p - easting_q) / distance**3


@jit(nopython=True)
def kernel_n(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Northing component of the gradient of the inverse of the distance

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_y(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial y}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        - \frac{
            y_p - y_q
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return -(northing_p - northing_q) / distance**3


@jit(nopython=True)
def kernel_u(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Upward component of the gradient of the inverse of the distance

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_z(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial z}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        - \frac{
            z_p - z_q
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return -(upward_p - upward_q) / distance**3


@jit(nopython=True)
def kernel_ee(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along easting-easting

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{xx}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial^2}{\partial x^2}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (x_p - x_q)^2
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }
        - \frac{
            1
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^3
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (easting_p - easting_q) ** 2 / distance**5 - 1 / distance**3


@jit(nopython=True)
def kernel_nn(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along northing-northing

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{yy}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial^2}{\partial y^2}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (y_p - y_q)^2
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }
        - \frac{
            1
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^3
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (northing_p - northing_q) ** 2 / distance**5 - 1 / distance**3


@jit(nopython=True)
def kernel_uu(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along upward-upward

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{zz}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial^2}{\partial z^2}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (z_p - z_q)^2
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }
        - \frac{
            1
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^3
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (upward_p - upward_q) ** 2 / distance**5 - 1 / distance**3


@jit(nopython=True)
def kernel_en(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along easting-northing

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{xy}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial x \partial y}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (x_p - x_q) (y_p - y_q)
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (easting_p - easting_q) * (northing_p - northing_q) / distance**5


@jit(nopython=True)
def kernel_eu(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along easting-upward

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{xz}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial x \partial z}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (x_p - x_q) (z_p - z_q)
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (easting_p - easting_q) * (upward_p - upward_q) / distance**5


@jit(nopython=True)
def kernel_nu(
    easting_p, northing_p, upward_p, easting_q, northing_q, upward_q, distance
):
    r"""
    Second derivative of the inverse of the distance along northing-upward

    .. important ::

        The coordinates of the two points must be in Cartesian coordinates and
        have the same units.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of point :math:`\mathbf{p}`.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of point :math:`\mathbf{q}`.
    distance : float
        Euclidean distance between points :math:`\mathbf{p}` and
        :math:`\mathbf{q}`.

    Returns
    -------
    kernel : float
        Value of the kernel function.

    Notes
    -----
    Given two points :math:`\mathbf{p} = (x_p, y_p, z_p)` and :math:`\mathbf{q}
    = (x_q, y_q, z_q)` defined in a Cartesian coordinate system, compute the
    following kernel function:

    .. math::

        k_{yz}(\mathbf{p}, \mathbf{q}) =
        \frac{\partial}{\partial y \partial z}
        \left(
            \frac{1}{\lVert \mathbf{p} - \mathbf{q} \rVert_2}
        \right)
        =
        \frac{
            3 (y_p - y_q) (z_p - z_q)
        }{
            \lVert \mathbf{p} - \mathbf{q} \rVert_2^5
        }

    where :math:`\lVert \cdot \rVert_2` refer to the :math:`L_2` norm (the
    Euclidean distance between :math:`\mathbf{p}` and :math:`\mathbf{q}`).
    """
    return 3 * (northing_p - northing_q) * (upward_p - upward_q) / distance**5



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Kernel functions for rectangular prisms
"""
import numpy as np
from numba import jit


@jit(nopython=True)
def kernel_pot(easting, northing, upward, radius):
    r"""
    Kernel for the potential field due to a rectangular prism

    Evaluates the integration kernel for the potential field generated by
    a prism [Nagy2000] on a single vertex of the prism. The coordinates that
    must be passed are shifted coordinates: the coordinates of the vertex from
    a Cartesian coordinate system whose origin is located in the observation
    point.

    This function makes use of a safe natural logarithmic function and a safe
    arctangent function [Fukushima2020]_ that guarantee a good accuracy on
    every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the numerical kernel function for the potential field due to
        a rectangular prism evaluated on a single vertex.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_V(x, y, z) &=
            x y \, \operatorname{safe\_ln} (z, r)
            + y z \, \operatorname{safe\_ln} (x, r)
            + z x \, \operatorname{safe\_ln} (y, r) \\
            & - \frac{x^2}{2} \operatorname{safe-arctan} \left( yz, xr \right)
            - \frac{y^2}{2} \operatorname{safe-arctan} \left( zx, yr \right)
            - \frac{z^2}{2} \operatorname{safe-arctan} \left( xy, zr \right)

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    and

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    kernel = (
        easting * northing * _safe_log(upward, easting, northing, radius)
        + northing * upward * _safe_log(easting, northing, upward, radius)
        + easting * upward * _safe_log(northing, upward, easting, radius)
        - 0.5 * easting**2 * _safe_atan2(upward * northing, easting * radius)
        - 0.5 * northing**2 * _safe_atan2(upward * easting, northing * radius)
        - 0.5 * upward**2 * _safe_atan2(easting * northing, upward * radius)
    )
    return kernel


@jit(nopython=True)
def kernel_e(easting, northing, upward, radius):
    r"""
    Kernel for easting component of the gradient due to a rectangular prism

    Evaluates the integration kernel for the easting component of the gradient
    of the potential field generated by a prism [Nagy2000]_ on a single vertex
    of the prism. The coordinates that must be passed are shifted coordinates:
    the coordinates of the vertex from a Cartesian coordinate system whose
    origin is located in the observation point.

    This function makes use of a safe natural logarithmic function and a safe
    arctangent function [Fukushima2020]_ that guarantee a good accuracy on
    every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the numerical kernel function for the easting component of the
        gradient of the potential field due to a rectangular prism evaluated on
        a single vertex.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_x(x, y, z) =
            -\left[
            y \, \operatorname{safe\_ln} (z, r)
            + z \, \operatorname{safe\_ln} (y, r)
            - x \, \operatorname{safe-arctan} \left( yz, xr \right)
            \right]

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    and

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    .. important::

        In the first equation a minus sign has been added to the one obtained
        by [Nagy2000]_ in order to compute the numerical kernel for the
        **eastward** component instead for the westward one.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    kernel = -(
        northing * _safe_log(upward, easting, northing, radius)
        + upward * _safe_log(northing, upward, easting, radius)
        - easting * _safe_atan2(northing * upward, easting * radius)
    )
    return kernel


@jit(nopython=True)
def kernel_n(easting, northing, upward, radius):
    r"""
    Kernel for northing component of the gradient due to a rectangular prism

    Evaluates the integration kernel for the northing component of the gradient
    of the potential field generated by a prism [Nagy2000]_ on a single vertex
    of the prism. The coordinates that must be passed are shifted coordinates:
    the coordinates of the vertex from a Cartesian coordinate system whose
    origin is located in the observation point.

    This function makes use of a safe natural logarithmic function and a safe
    arctangent function [Fukushima2020]_ that guarantee a good accuracy on
    every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the numerical kernel function for the northing component of
        the gradient of the potential field due to a rectangular prism
        evaluated on a single vertex.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_y(x, y, z) =
            -\left[
            z \, \operatorname{safe\_ln} (x, r)
            + x \, \operatorname{safe\_ln} (z, r)
            - y \, \operatorname{safe-arctan} \left( zx, yr \right)
            \right]

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    and

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    .. important::

        In the first equation a minus sign has been added to the one obtained
        by [Nagy2000]_ in order to compute the numerical kernel for the
        **northward** component instead for the southward one.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    kernel = -(
        upward * _safe_log(easting, northing, upward, radius)
        + easting * _safe_log(upward, easting, northing, radius)
        - northing * _safe_atan2(easting * upward, northing * radius)
    )
    return kernel


@jit(nopython=True)
def kernel_u(easting, northing, upward, radius):
    r"""
    Kernel for upward component of the gradient due to a rectangular prism

    Evaluates the integration kernel for the upward component of the gradient
    of the potential field generated by a prism [Nagy2000]_ on a single vertex
    of the prism. The coordinates that must be passed are shifted coordinates:
    the coordinates of the vertex from a Cartesian coordinate system whose
    origin is located in the observation point.

    This function makes use of a safe natural logarithmic function and a safe
    arctangent function [Fukushima2020]_ that guarantee a good accuracy on
    every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the numerical kernel function for the upward component of the
        gradient of the potential field due to a rectangular prism evaluated on
        a single vertex.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_z(x, y, z) =
            - \left[
            x \, \operatorname{safe\_ln} (y, r)
            + y \, \operatorname{safe\_ln} (x, r)
            - z \, \operatorname{safe-arctan} \left( xy, zr \right)
            \right]

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    and

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    .. important::

        In the first equation a minus sign has been added to the one obtained
        by [Nagy2000]_ in order to compute the numerical kernel for the
        **upward** component instead for the downward one.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # The minus sign is to return the kernel for the upward component instead
    # of the downward one.
    kernel = -(
        easting * _safe_log(northing, upward, easting, radius)
        + northing * _safe_log(easting, northing, upward, radius)
        - upward * _safe_atan2(easting * northing, upward * radius)
    )
    return kernel


@jit(nopython=True)
def kernel_ee(easting, northing, upward, radius):
    r"""
    Kernel for easting-easting component of the tensor due to a prism

    Evaluates the integration kernel for the easting-easting component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe arctangent function [Fukushima2020]_ that
    guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-easting component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xx}(x, y, z) = - \operatorname{safe-arctan} \left( yz, xr \right)

    where

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return -_safe_atan2(northing * upward, easting * radius)


@jit(nopython=True)
def kernel_nn(easting, northing, upward, radius):
    r"""
    Kernel for northing-northing component of the tensor due to a prism

    Evaluates the integration kernel for the northing-northing component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe arctangent function [Fukushima2020]_ that
    guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the northing-northing component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{yy}(x, y, z) = - \operatorname{safe-arctan} \left( zx, yr \right)

    where

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return -_safe_atan2(easting * upward, northing * radius)


@jit(nopython=True)
def kernel_uu(easting, northing, upward, radius):
    r"""
    Kernel for upward-upward component of the tensor due to a prism

    Evaluates the integration kernel for the upward-upward component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe arctangent function [Fukushima2020]_ that
    guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the upward-upward component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{zz}(x, y, z) = - \operatorname{safe-arctan} \left( xy, zr \right)

    where

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return -_safe_atan2(easting * northing, upward * radius)


@jit(nopython=True)
def kernel_en(easting, northing, upward, radius):
    r"""
    Kernel for easting-northing component of the tensor due to a prism

    Evaluates the integration kernel for the easting-northing component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe natural logarithmic function
    [Fukushima2020]_ that guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-northing component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xy}(x, y, z) = \operatorname{safe\_ln} \left(z, r \right)

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return _safe_log(upward, easting, northing, radius)


@jit(nopython=True)
def kernel_eu(easting, northing, upward, radius):
    r"""
    Kernel for easting-upward component of the tensor due to a prism

    Evaluates the integration kernel for the easting-upward component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe natural logarithmic function
    [Fukushima2020]_ that guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-upward component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xz}(x, y, z) = \operatorname{safe\_ln} \left(y, r \right)

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return _safe_log(northing, easting, upward, radius)


@jit(nopython=True)
def kernel_nu(easting, northing, upward, radius):
    r"""
    Kernel for northing-upward component of the tensor due to a prism

    Evaluates the integration kernel for the northing-upward component of the
    tensor generated by a prism [Nagy2000]_ on a single vertex of the prism.
    The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    This function makes use of a safe natural logarithmic function
    [Fukushima2020]_ that guarantee a good accuracy on every observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the northing-upward component of the
        tensor due to a rectangular prism evaluated on a single vertex.
        Return ``np.nan`` if ``radius`` is zero.

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{yz}(x, y, z) = \operatorname{safe\_ln} \left(x, r \right)

    where

    .. math::

        \operatorname{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    if radius == 0.0:
        return np.nan
    return _safe_log(easting, northing, upward, radius)


@jit(nopython=True)
def _safe_atan2(y, x):
    r"""
    Principal value of the arctangent expressed as a two variable function

    This modification has to be made to the arctangent function so the
    gravitational field of the prism satisfies the Poisson's equation.
    Therefore, it guarantees that the fields satisfies the symmetry properties
    of the prism. This modified function has been defined according to
    [Fukushima2020]_.

    Notes
    -----

    .. math::

        \operatorname{safe-arctan} \left(y, x\right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    References
    ----------
    - [Fukushima2020]_
    """
    if x == 0:
        if y > 0:
            result = np.pi / 2
        elif y < 0:
            result = -np.pi / 2
        else:
            result = 0
        return result
    return np.arctan(y / x)


@jit(nopython=True)
def _safe_log(x, y, z, r):
    r"""
    Safe log function to use in the prism kernels.

    Evaluates the :math:`\ln(x + r)` where :math:`x` is one of the shifted
    coordinate of the prism vertex and :math:`r` is the Euclidean distance
    (always non-negative) from the vertex to the observation point.

    Parameters
    ----------
    x : float
        Shifted coordinate of the prism vertex that will be used for evaluating
        the log.
    y, z : float
        The other two shifted coordinates of the prism vertex. They are used to
        determine if ``abs(x) == r`` with more accuracy, and to evaluate the
        log function to avoid floating point errors when :math:`x < 0` and
        :math:`r \ne |x|`.
    r : float
        Euclidean distance from the vertex to the observation point. We require
        this argument because this quantity is usually precomputed before
        evaluating the kernel, thus saving computation time.

    Returns
    -------
    float
        Evaluation of the :math:`\ln{x + r}` to be used on kernels.

    Notes
    -----
    The :math:`\text{safe\_ln}(x, r)` function is defined as:

    .. math::

        \text{safe\_ln}(x, r) =
        \begin{cases}
            0 & r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((y^2 + z^2) / (r - x)) & x < 0, r \ne |x| \\
            -\ln(-2 x) & x < 0, r = |x|
        \end{cases}

    The function evaluates :math:`\ln(x + r)` when :math:`x \ge 0`. Otherwise,
    the returned value takes into account the limit cases:

    * If :math:`r = 0` (observation point is on one of the vertices of the
      prism), it returns zero. This accounts for the limits of the
      zeroth-order and first-order kernels when evaluated on the vertices of
      the prism. The second-order kernels are not defined on the vertices.
    * If :math:`x < 0, r \ne |x|`, then it evaluates
      :math:`\ln((y^2 + z^2) / (r - x))` to reduce floating point errors
      ([Fukushima2020]_).
    * If :math:`x < 0, r = |x|`, then it returns one of the terms of the limit
      of the second-order kernels (when the observation point is inline with
      two nodes): :math:`-\ln(-2x)`.

    When checking if :math:`r = |x|`, we will evaluate if ``y == 0.0 and z ==
    0.0``, to avoid issues with floating point errors in cases in which
    :math:`|x| \gg |y|` and/or  :math:`|x| \gg |z|`. In such cases, the ``r``
    value could be exactly the same as ``x`` (up to machine precision) and
    not reflect that ``y`` and/or ``z`` are not-null.

    This modified version of the log function was inspired by [Nagy2000]_ and
    [Fukushima2020]_.

    References
    ----------
    - [Nagy2000]_
    - [Fukushima2020]_
    """
    if r == 0:
        return 0
    if x < 0:
        if y == 0.0 and z == 0.0:
            result = -np.log(-2 * x)
        else:
            result = np.log((y**2 + z**2) / (r - x))
        return result
    return np.log(x + r)


@jit(nopython=True)
def kernel_eee(easting, northing, upward, radius):
    r"""
    Easting-easting-easting kernel due to a prism.

    Evaluates the integration kernel for the easting-easting-easting component
    of the grad tensor generated by a prism [Nagy2000]_ on a single vertex of
    the prism. The coordinates that must be passed are shifted coordinates: the
    coordinates of the vertex from a Cartesian coordinate system whose origin
    is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-easting-easting component
        of the grad tensor due to a rectangular prism evaluated on a single
        vertex. Return ``np.nan`` if ``radius`` is zero (i.e. observation point
        is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xxx}(x, y, z) =
            \frac{
                - y z (2 x^2 + y^2 + z^2)
            }{
                (x^2 + y^2)(x^2 + z^2)
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_eee(x, y, z, r)) # doctest: +NUMBER
    0.18706595
    """
    return _kernel_iii(easting, northing, upward, radius)


@jit(nopython=True)
def kernel_nnn(easting, northing, upward, radius):
    r"""
    Northing-northing-northing kernel due to a prism.

    Evaluates the integration kernel for the northing-northing-northing
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the northing-northing-northing
        component of the grad tensor due to a rectangular prism evaluated on
        a single vertex. Return ``np.nan`` if ``radius`` is zero (i.e.
        observation point is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{yyy}(x, y, z) =
            \frac{
                - x z (x^2 + 2 y^2 + z^2)
            }{
                (x^2 + y^2)(y^2 + z^2)
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_nnn(x, y, z, r)) # doctest: +NUMBER
    0.07574927
    """
    return _kernel_iii(northing, upward, easting, radius)


@jit(nopython=True)
def kernel_uuu(easting, northing, upward, radius):
    r"""
    Upward-upward-upward kernel due to a prism.

    Evaluates the integration kernel for the upward-upward-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the upward-upward-upward component of
        the grad tensor due to a rectangular prism evaluated on a single
        vertex. Return ``np.nan`` if ``radius`` is zero (i.e. observation point
        is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{zzz}(x, y, z) =
            \frac{
                - x y (x^2 + y^2 + 2 z^2)
            }{
                (x^2 + z^2)(y^2 + z^2)
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_uuu(x, y, z, r)) # doctest: +NUMBER
    -0.19440331
    """
    return _kernel_iii(upward, northing, easting, radius)


@jit(nopython=True)
def kernel_een(easting, northing, upward, radius):
    r"""
    Easting-easting-northing kernel due to a prism.

    Evaluates the integration kernel for the easting-easting-northing
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-easting-northing component
        of the grad tensor due to a rectangular prism evaluated on a single
        vertex. Return ``np.nan`` if ``radius`` is zero (i.e. observation point
        is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xxy}(x, y, z) =
            \frac{
                - x
            }{
                (z + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_een(x, y, z, r)) # doctest: +NUMBER
    -0.12214070
    """
    return _kernel_iij(easting, northing, upward, radius)


@jit(nopython=True)
def kernel_eeu(easting, northing, upward, radius):
    r"""
    Easting-easting-upward kernel due to a prism.

    Evaluates the integration kernel for the easting-easting-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-easting-upward component
        of the grad tensor due to a rectangular prism evaluated on a single
        vertex. Return ``np.nan`` if ``radius`` is zero (i.e. observation point
        is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xxz}(x, y, z) =
            \frac{
                - x
            }{
                (y + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_eeu(x, y, z, r)) # doctest: +NUMBER
    -0.03837408
    """
    return _kernel_iij(easting, upward, northing, radius)


@jit(nopython=True)
def kernel_enn(easting, northing, upward, radius):
    r"""
    Easting-northing-northing kernel due to a prism.

    Evaluates the integration kernel for the easting-northing-northing
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-northing-northing
        component of the grad tensor due to a rectangular prism evaluated on
        a single vertex. Return ``np.nan`` if ``radius`` is zero (i.e.
        observation point is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xyy}(x, y, z) =
            \frac{
                - y
            }{
                (z + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_enn(x, y, z, r)) # doctest: +NUMBER
    -0.20488118
    """
    return _kernel_iij(northing, easting, upward, radius)


@jit(nopython=True)
def kernel_nnu(easting, northing, upward, radius):
    r"""
    Northing-northing-upward kernel due to a prism.

    Evaluates the integration kernel for the northing-northing-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the northing-northing-upward
        component of the grad tensor due to a rectangular prism evaluated on
        a single vertex. Return ``np.nan`` if ``radius`` is zero (i.e.
        observation point is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{yyz}(x, y, z) =
            \frac{
                - y
            }{
                (x + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_nnu(x, y, z, r)) # doctest: +NUMBER
    -0.07808384
    """
    return _kernel_iij(northing, upward, easting, radius)


@jit(nopython=True)
def kernel_euu(easting, northing, upward, radius):
    r"""
    Easting-upward-upward kernel due to a prism.

    Evaluates the integration kernel for the easting-upward-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-upward-upward
        component of the grad tensor due to a rectangular prism evaluated on
        a single vertex. Return ``np.nan`` if ``radius`` is zero (i.e.
        observation point is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xzz}(x, y, z) =
            \frac{
                - z
            }{
                (y + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_euu(x, y, z, r)) # doctest: +NUMBER
    0.03713621
    """
    return _kernel_iij(upward, easting, northing, radius)


@jit(nopython=True)
def kernel_nuu(easting, northing, upward, radius):
    r"""
    Northing-upward-upward kernel due to a prism.

    Evaluates the integration kernel for the northing-upward-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the northing-upward-upward
        component of the grad tensor due to a rectangular prism evaluated on
        a single vertex. Return ``np.nan`` if ``radius`` is zero (i.e.
        observation point is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{yzz}(x, y, z) =
            \frac{
                - z
            }{
                (x + \sqrt{x^2 + y^2 + z^2})
                \sqrt{x^2 + y^2 + z^2}
            }

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_nuu(x, y, z, r)) # doctest: +NUMBER
    0.04504837
    """
    return _kernel_iij(upward, northing, easting, radius)


@jit(nopython=True)
def kernel_enu(easting, northing, upward, radius):
    r"""
    Easting-northing-upward kernel due to a prism.

    Evaluates the integration kernel for the easting-northing-upward
    component of the grad tensor generated by a prism [Nagy2000]_ on a single
    vertex of the prism. The coordinates that must be passed are shifted
    coordinates: the coordinates of the vertex from a Cartesian coordinate
    system whose origin is located in the observation point.

    Parameters
    ----------
    easting, northing, upward : float
        Shifted easting, northing and upward coordinates of the vertex of the
        prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``easting``, ``northing``
        and ``upward`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function for the easting-northing-upward component
        of the grad tensor due to a rectangular prism evaluated on a single
        vertex. Return ``np.nan`` if ``radius`` is zero (i.e. observation point
        is on the vertex).

    Notes
    -----
    Computes the following numerical kernel on the passed *shifted
    coordinates*:

    .. math::

        k_{xyz}(x, y, z) = \frac{-1}{\sqrt{x^2 + y^2 + z^2}}

    Examples
    --------
    >>> x, y, z = 3.1, 5.2, -3.0
    >>> r = np.sqrt(x**2 + y**2 + z**2)
    >>> float(kernel_enu(x, y, z, r)) # doctest: +NUMBER
    -0.1480061
    """
    if radius == 0.0:
        return np.nan
    return -1 / radius


@jit(nopython=True)
def _kernel_iii(x_i, x_j, x_k, radius):
    r"""
    Diagonal 3rd order kernels of a prism.

    Evaluates the following integration kernel:

    .. math::

        k_{iii}(x_i, x_j, x_k) =
            \frac{
                - x_j x_k (2 x_i^2 + x_j^2 + x_k^2)
            }{
                (x_i^2 + x_j^2)(x_i^2 + x_k^2)
                \sqrt{x_i^2 + x_j^2 + x_k^2}
            }

    This function allows us to evaluate the third order kernels
    :math:`k_{xxx}(x, y, z)`, :math:`k_{yyy}(x, y, z)`, and :math:`k_{zzz}(x,
    y, z)` by cycling through the coordinates of the nodes of the prism.

    Parameters
    ----------
    x_i, x_j, x_k : float
        Shifted coordinates of the vertex of the prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``x_i``, ``x_j``, and
        ``x_k`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function. Return ``np.nan`` if ``radius`` is zero.

    Examples
    --------
    Given the shifted coordinates of a prism vertex ``easting``, ``northing``
    and ``upward``, we can use this function to compute the diagonal third
    order kernels by cycling the order of the shifted coordinates of the
    vertex:

    >>> import numpy as np
    >>> easting, northing, upward = 1.7, 2.8, 3.9
    >>> radius = np.sqrt(easting**2 + northing**2 + upward**2)
    >>> k_eee = _kernel_iii(easting, northing, upward, radius)
    >>> k_nnn = _kernel_iii(northing, upward, easting, radius)
    >>> k_uuu = _kernel_iii(upward, easting, northing, radius)
    """
    if radius == 0.0:
        return np.nan
    if (x_i == 0 and x_j == 0) or (x_i == 0 and x_k == 0):
        return 0.0
    x_i_sq, x_j_sq, x_k_sq = x_i**2, x_j**2, x_k**2
    numerator = -x_j * x_k * (2 * x_i_sq + x_j_sq + x_k_sq)
    denominator = (x_i_sq + x_j_sq) * (x_i_sq + x_k_sq) * radius
    return numerator / denominator


@jit(nopython=True)
def _kernel_iij(x_i, x_j, x_k, radius):
    r"""
    Non-diagonal 3rd order kernels of a prism.

    Evaluates the following integration kernel:

    .. math::

        easting / ((upward + radius) * radius)

        k_{iij}(x_i, x_j, x_k) =
            \frac{
                - x_i
            }{
                (x_k + \sqrt{x_i^2 + x_j^2 + x_k^2})
                \sqrt{x_i^2 + x_j^2 + x_k^2}
            }

    This function allows us to evaluate the non-diagonal third order kernels
    :math:`k_{xxy}(x, y, z)`, :math:`k_{xxz}(x, y, z)`,
    :math:`k_{xyy}(x, y, z)`, :math:`k_{yyz}(x, y, z)`,
    :math:`k_{xzz}(x, y, z)`, and :math:`k_{yzz}(x, y, z)` by cycling
    through the coordinates of the nodes of the prism.

    Parameters
    ----------
    x_i, x_j, x_k : float
        Shifted coordinates of the vertex of the prism. Must be in meters.
    radius : float
        Square root of the sum of the squares of the ``x_i``, ``x_j``, and
        ``x_k`` shifted coordinates.

    Returns
    -------
    kernel : float
        Value of the kernel function. Return ``np.nan`` if ``radius`` is zero.

    Examples
    --------
    Given the shifted coordinates of a prism vertex ``easting``, ``northing``
    and ``upward``, we can use this function to compute the non-diagonal third
    order kernels by changing the order of the shifted coordinates of the
    vertex:

    >>> import numpy as np
    >>> easting, northing, upward = 1.7, 2.8, 3.9
    >>> radius = np.sqrt(easting**2 + northing**2 + upward**2)
    >>> k_een = _kernel_iij(easting, northing, upward, radius)
    >>> k_eeu = _kernel_iij(easting, upward, northing, radius)
    >>> k_enn = _kernel_iij(northing, easting, upward, radius)
    >>> k_nnu = _kernel_iij(northing, upward, easting, radius)
    >>> k_euu = _kernel_iij(upward, easting, northing, radius)
    >>> k_nuu = _kernel_iij(upward, northing, upward, radius)
    """
    if radius == 0.0:
        return np.nan
    if x_i == 0 and x_j == 0:
        return 0.0
    return -x_i / ((x_k + radius) * radius)



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Test the _safe_log private function for prisms.
"""

import numpy as np
import numpy.testing as npt
import pytest

from ..prism._kernels import _safe_log


class TestSafeLogFixedX:
    r"""
    Test some properties of ``_safe_log`` when ``x`` is fixed.

    We define:

    .. math::

        f(x, y, z) = \ln(x + \sqrt{x^2 + y^2 + z^2})

    When fixing ``x``, this function has some continuity, symmetry and monotony
    properties.

    For simplicity, let's define:

    .. math::

        g(z) = f(x=x_0, y=0, z) = \ln(x_0 + \sqrt{x_0^2 + z^2})

    with :math:`x_0 = \text{constant}`.
    """

    def eval_safe_log(self, z, x_0=0.0):
        r = np.sqrt(x_0**2 + z**2)
        return _safe_log(x_0, 0.0, z, r)

    @pytest.mark.parametrize(
        "x_0", (-10.0, 0.0, 10.0), ids=("x_0=-10.0", "x_0=0.0", "x_0=10.0")
    )
    def test_monotony_z_negative(self, x_0):
        """
        Test monotony of the safe_log function with negative z values.

        Function :math:`g(x, z)` with :math:`x` fixed monotonically decreases
        on :math:`z < 0`. Without fixes to reduce the floating point errors,
        the safe_log function won't be monotone.
        """
        z = np.linspace(-1e-3, 0, 1001)
        z = z[z < 0]
        results = np.array([self.eval_safe_log(z_i, x_0) for z_i in z])
        assert (results[1:] < results[:-1]).all()

    @pytest.mark.parametrize(
        "x_0", (-10.0, 0.0, 10.0), ids=("x_0=-10.0", "x_0=0.0", "x_0=10.0")
    )
    def test_monotony_z_positive(self, x_0):
        """
        Test monotony of the safe_log function with positive z values.

        Function :math:`g(x, z)` with :math:`x` fixed monotonically increases
        on :math:`z > 0`. Without fixes to reduce the floating point errors,
        the safe_log function won't be monotone.
        """
        z = np.linspace(0, 1e-3, 1001)
        z = z[z > 0]
        results = np.array([self.eval_safe_log(z_i, x_0) for z_i in z])
        assert (results[1:] > results[:-1]).all()

    @pytest.mark.parametrize(
        "x_0", (-10.0, 0.0, 10.0), ids=("x_0=-10.0", "x_0=0.0", "x_0=10.0")
    )
    def test_symmetry(self, x_0):
        """
        Test symmetry of the function with respect to zero.

        Function :math:`g(x, z)` with :math:`x` fixed is symmetric with respect
        to :math:`z = 0`.
        """
        vmax = 1e-3
        z = np.linspace(-vmax, vmax, 1001)
        z = z[z != 0]
        results = np.array([self.eval_safe_log(z_i, x_0) for z_i in z])
        results_z_negative = results[z < 0]
        results_z_positive = results[z > 0]
        npt.assert_allclose(results_z_negative, results_z_positive[::-1])



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Define utility functions for running the tests
"""
import os

import numpy as np

# Determine if Numba just-in-time compilation is disabled
NUMBA_IS_DISABLED = os.environ.get("NUMBA_DISABLE_JIT", default="0") != "0"


def spherical_to_geocentric_cartesian(longitude, latitude, radius):
    """
    Convert spherical coordinates into geocentric Cartesian coordinates

    Parameters
    ----------
    longitude : float
        Longitudinal coordinate (in degrees).
    latitude : float
        Latitudinal coordinate (in degrees).
    radius : float
        Radial coordinates (in meters).

    Returns
    -------
    x, y, z : floats
        Geocentric Cartesian coordinates of the passed point.
    """
    longitude, latitude = np.radians(longitude), np.radians(latitude)
    x = radius * np.cos(longitude) * np.cos(latitude)
    y = radius * np.sin(longitude) * np.cos(latitude)
    z = radius * np.sin(latitude)
    return x, y, z


def dumb_spherical_distance(point_p, point_q):
    """
    Dumb calculation of distance between two points in spherical coordinates

    This function converts the spherical coordinates of each point to
    geocentric Cartesian coordinates and then uses a Cartesian Euclidean
    distance calculation.

    .. warning:

        This implementation is less optimal than the one we have in
        :func:`choclo.distance_spherical`. It exists only for testing purposes.

    .. important:

        All angles must be in degrees and radii in meters.

    Parameters
    ----------
    point_p : tuple or 1d-array
        Tuple or array containing the coordinates of the first point in the
        following order: (``longitude``, ``latitude`` and ``radius``).
        Both ``longitude`` and ``latitude`` must be in degrees, while
        ``radius`` in meters.
    point_q : tuple or 1d-array
        Tuple or array containing the coordinates of the second point in the
        following order: (``longitude``, ``latitude`` and ``radius``).
        Both ``longitude`` and ``latitude`` must be in degrees, while
        ``radius`` in meters.

    Returns
    -------
    distance : float
        Euclidean distance between ``point_p`` and ``point_q``.
    """
    # Convert points to geocentric Cartesian coordinates
    point_p = spherical_to_geocentric_cartesian(*point_p)
    point_q = spherical_to_geocentric_cartesian(*point_q)
    # Convert points to arrays
    point_p, point_q = np.asarray(point_p), np.asarray(point_q)
    # Calculate Euclidean Cartesian distance between these two
    return np.sqrt(np.sum((point_p - point_q) ** 2))
